/*****************************************************
Chip type               : ATmega8L
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/
#include <mega8.h>
#include <math.h>
#include <delay.h>

#define ticks_per_half_wave 156          
#define ticks_per_wave (ticks_per_half_wave*2)  
#define pwm_top 0x01FF
#define sd_pin PORTB.0
#define ocp_led PORTB.4
#define battery_low_led PORTB.5
#define battery_min_voltage 155 // 11V
#define AC_normal_val 227 //310V



volatile unsigned int sin_table [ticks_per_wave];
volatile unsigned int i = 0;

volatile unsigned char timer_ticks;
volatile unsigned char U_bat;
volatile unsigned char timer_ticks_2;
volatile bit U_good; 
volatile unsigned char U_AC;
volatile unsigned char k;

void sin_init (void) 
{ 
 unsigned int n = 0; 
 unsigned int f = 0; 
 unsigned int c = 0; 
 float g= (float) PI / ticks_per_half_wave;

 do 
    { 
    if (n <= ticks_per_half_wave)
    {sin_table[n]=(unsigned int)((pwm_top>>1) + (sin((float) (f * g)) * ((pwm_top-(pwm_top/3))/2))); f++;}
    else  {sin_table[n]=(unsigned int)((pwm_top>>1) - (sin((float) (c * g))* ((pwm_top-(pwm_top/3))/2))); c++;}
    
    n++; 
    } while (n<(ticks_per_wave+1));

 }  
//___________________________________________________________________________________
void port_init ()
{
DDRB = 0b00110011;
PORTB = 0b00000000;

DDRD = 0b00000000;
PORTD = 0b00000000;

DDRC = 0b00000000;
PORTC = 0b00000000;
}
//___________________________________________________________________________________
void ACMP_init () 
{ACSR |= (1<<ACBG);} 
//___________________________________________________________________________________
void ADC_init()
{
  ADMUX |= (1<<REFS1)|(1<<REFS0)|(1<<ADLAR) ;
  ADCSRA |= (1<<ADEN)|(1<<ADPS2)/*|(1<<ADPS1)*/|(1<<ADPS0);
}
//___________________________________________________________________________________
void TCCR0_init ()
{
TCCR0=0x05;
TIMSK |= (1<<TOIE0);
}
//___________________________________________________________________________________
void TCCR1_init ()
{
OCR1A = pwm_top>>1; 
OCR1B = pwm_top>>1;
TCCR1A = 0b10000000;
TCCR1B = 0b00000001;
TCCR1A |= 1<<WGM11;
TCCR1B |= 1<<WGM12;

TIMSK |= 1<<OCIE1A;
}
//___________________________________________________________________________________
interrupt [TIM1_COMPA] void timer1_compa_isr(void)
{
if  (i == ticks_per_wave) {i=0;}

            if (i <= ticks_per_half_wave)
                     { OCR1A =  sin_table [i] + k;}
            else  { OCR1A =  sin_table [i] -  k;}         


        
  i++;       
}
//___________________________________________________________________________________
void ADC_read ()
{
static unsigned int tmp_U = 0;
static unsigned char i_time = 255;
static bit t_bit = 0; 
 
 if ((ADCSRA.ADSC == 0)&&(t_bit))
    {
    if  (i_time) {tmp_U += ADCH; i_time--;}
    else 
        {
        U_bat = (tmp_U >> 8); tmp_U = 0; i_time = 255; 
        t_bit = 0;
        ADMUX.MUX0 = 0;
        ADMUX.MUX1 = 0;
        ADMUX.MUX2 = 0;
        }
   ADCSRA |= (1<<ADSC);
 }
 else if (ADCSRA.ADSC == 0)
 {
  if  (i_time) {tmp_U += ADCH; i_time--;}
  else 
        {
        U_AC = (tmp_U >> 8); tmp_U = 0; i_time = 255; 
        t_bit = 1;
        ADMUX.MUX0 = 1;
        ADMUX.MUX1 = 1;
        ADMUX.MUX2 = 1;
        }
   ADCSRA |= (1<<ADSC);
}
}

//___________________________________________________________________________________
void battery_check ()
{
if (U_bat < battery_min_voltage) {sd_pin = 0; battery_low_led = 1; U_good = 0;}
else if (U_bat > (battery_min_voltage + 20))
{ 
if (U_good == 0) {timer_ticks_2 = 30; U_good = 1;}
if ((timer_ticks_2 == 0)&&(timer_ticks == 0)) {sd_pin = 1; battery_low_led = 0;} 
}
}
//___________________________________________________________________________________
interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
if (timer_ticks) {timer_ticks -- ;}
if (timer_ticks_2) {timer_ticks_2 -- ;}
battery_check ();
}
//___________________________________________________________________________________
void ocp ()
{
if (ACSR.ACO == 0) {timer_ticks = 60;}
if (timer_ticks) {sd_pin = 0; ocp_led = 1;}
else if ((U_good)&&(timer_ticks_2 == 0)) {sd_pin = 1; ocp_led= 0;}
else {ocp_led= 0;}
}
//___________________________________________________________________________________
void korr_AC ()
{
    unsigned int tmp;
    if (U_AC < AC_normal_val) 
    { 
    tmp =  (unsigned int)((pwm_top>>1)*(float) (AC_normal_val - U_AC) / AC_normal_val) ;
    if ( (sin_table [(ticks_per_half_wave>>1)] + tmp) < (pwm_top - (pwm_top>>4)) ) { k = tmp;}
    }
    
    else  if ((U_AC > AC_normal_val)&&(k)) { k -- ; }
}                                                                                                      
//___________________________________________________________________________________



void main(void)
{
port_init ();
sin_init ();
ADC_init ();
ACMP_init ();
TCCR0_init ();
TCCR1_init ();

timer_ticks = 30;
U_AC = AC_normal_val;

#asm("sei")

while (1) {ADC_read (); ocp (); korr_AC ();}

}
